"""
This module implements the ISMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['ISMOTE']

class ISMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{ismote,
                            author="Li, Hu
                            and Zou, Peng
                            and Wang, Xiang
                            and Xia, Rongze",
                            editor="Sun, Zengqi
                            and Deng, Zhidong",
                            title="A New Combination Sampling Method for
                                    Imbalanced Data",
                            booktitle="Proceedings of 2013 Chinese Intelligent
                                        Automation Conference",
                            year="2013",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="547--554",
                            isbn="978-3-642-38466-0"
                            }
    """

    categories = [OverSamplingSimplex.cat_changes_majority,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 minority_weight=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            minority_weight (float): weight parameter according to the paper
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(minority_weight, "minority_weight", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.minority_weight = minority_weight
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'n_neighbors': [3, 5, 7],
                                  'minority_weight': [0.2, 0.5, 0.8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        n_to_sample = int((len(X_maj) - len(X_min))/2 + 0.5)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # computing distances of majority samples from minority ones
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=len(X_min),
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        dist, ind = nnmt.kneighbors(X_maj)

        # sort majority instances in descending order by their mean distance
        # from minority samples
        to_sort = zip(np.arange(len(X_maj)), np.mean(dist, axis=1))
        ind_sorted, _ = zip(*sorted(to_sort, key=lambda x: -x[1]))

        # remove the ones being farthest from the minority samples
        X_maj = X_maj[list(ind_sorted[n_to_sample:])]

        # construct new dataset
        X = np.vstack([X_maj, X_min])
        y = np.hstack([np.repeat(self.maj_label, len(X_maj)),
                           np.repeat(self.min_label, len(X_min))])

        # fitting nearest neighbors model
        n_neighbors = np.min([len(X), self.n_neighbors + 1])
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        vertex_weights = np.repeat(1.0 - self.minority_weight, X.shape[0])
        vertex_weights[y == self.min_label] = self.minority_weight

        # removing the minority samples from their own neighborhoods
        ind = ind[:, 1:]

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        X_vertices=X,
                                        vertex_weights=vertex_weights)

        # do the oversampling
        #samples = []
        #while len(samples) < n_to_sample:
        #    idx = self.random_state.choice(np.arange(len(X_min)))
        #    y_idx = self.random_state.choice(ind[idx][1:])

            # different generation scheme depending on the class label
        #    if y_new[y_idx] == self.min_label:
        #        diff = (X_new[y_idx] - X_min[idx])
        #        r = self.random_state.random_sample()
        #        samples.append(X_min[idx] + r * diff * self.minority_weight)
        #    else:
        #        diff = (X_new[y_idx] - X_min[idx])
        #        r = self.random_state.random_sample()
        #        sample = X_min[idx] + r * diff * (1.0 - self.minority_weight)
        #        samples.append(sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'minority_weight': self.minority_weight,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Assembled_SMOTE method.
"""
import warnings

import numpy as np

from sklearn.decomposition import PCA

from ..config import suppress_external_warnings
from ..base import coalesce_dict, coalesce
from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['Assembled_SMOTE']

class Assembled_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{assembled_smote,
                            author={Zhou, B. and Yang, C. and Guo, H. and
                                        Hu, J.},
                            booktitle={The 2013 International Joint Conference
                                        on Neural Networks (IJCNN)},
                            title={A quasi-linear SVM combined with assembled
                                    SMOTE for imbalanced data classification},
                            year={2013},
                            volume={},
                            number={},
                            pages={1-7},
                            keywords={approximation theory;interpolation;
                                        pattern classification;sampling
                                        methods;support vector machines;trees
                                        (mathematics);quasilinear SVM;
                                        assembled SMOTE;imbalanced dataset
                                        classification problem;oversampling
                                        method;quasilinear kernel function;
                                        approximate nonlinear separation
                                        boundary;mulitlocal linear boundaries;
                                        interpolation;data distribution
                                        information;minimal spanning tree;
                                        local linear partitioning method;
                                        linear separation boundary;synthetic
                                        minority class samples;oversampled
                                        dataset classification;standard SVM;
                                        composite quasilinear kernel function;
                                        artificial data datasets;benchmark
                                        datasets;classification performance
                                        improvement;synthetic minority
                                        over-sampling technique;Support vector
                                        machines;Kernel;Merging;Standards;
                                        Sociology;Statistics;Interpolation},
                            doi={10.1109/IJCNN.2013.6707035},
                            ISSN={2161-4407},
                            month={Aug}}

    Notes:
        * Absolute value of the angles extracted should be taken.
            (implemented this way)
        * It is not specified how many samples are generated in the various
            clusters.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 pop=2,
                 thres=0.3,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            pop (int): lower threshold on cluster sizes
            thres (float): threshold on angles
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(pop, "pop", 1)
        self.check_in_range(thres, "thres", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.pop = pop
        self.thres = thres
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'pop': [2, 4, 5],
                                  'thres': [0.1, 0.3, 0.5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_border_non_border(self, X, y, X_min, nn_params):
        """
        Determine the border and non-border samples

        Args:
            X (np.array): all samples
            y (np.array): the target labels
            X_min (np.array): the minority samples
            nn_params (dict): the parameters of nearest neighbors

        Returns:
            np.array, np.array: the border and non-border samples
        """
        # fitting nearest neighbors model
        n_neighbors = min([len(X), self.n_neighbors+1])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nearestn.fit(X)
        ind = nearestn.kneighbors(X_min, return_distance=False)

        # finding the set of border and non-border minority elements
        n_min_neighbors = [np.sum(y[ind[i]] == self.min_label)
                           for i in range(len(ind))]
        border_mask = np.logical_not(np.array(n_min_neighbors) == n_neighbors)
        X_border = X_min[border_mask] # pylint: disable=invalid-name
        X_non_border = X_min[np.logical_not(border_mask)] # pylint: disable=invalid-name

        return X_border, X_non_border

    def do_the_clustering(self,
                            X_border,  # pylint: disable=invalid-name
                            nn_params):
        """
        Carry out the clustering of the border samples.

        Args:
            X_border (np.array): the border samples
            nn_params (dict): the nearest neighbor parameters

        Returns:
            list(np.array): the clusters
        """

        # initializing clustering
        clusters = [np.array([i]) for i in range(len(X_border))]

        distm = pairwise_distances_mahalanobis(X_border,
                                            tensor=nn_params.get('metric_tensor', None))
        for idx in range(len(distm)):
            distm[idx, idx] = np.inf

        # do the clustering
        while len(distm) > 1 and np.min(distm) < np.inf:
            # extracting coordinates of clusters with the minimum distance
            min_coord = np.where(distm == np.min(distm))
            merge_a = min_coord[0][0]
            merge_b = min_coord[1][0]

            # checking the size of clusters to see if they should be merged
            if (len(clusters[merge_a]) < self.pop
                    or len(clusters[merge_b]) < self.pop):
                # if both clusters are small, do the merge
                clusters[merge_a] = np.hstack([clusters[merge_a],
                                               clusters[merge_b]])
                del clusters[merge_b]
                # update the distance matrix accordingly
                distm[merge_a] = np.min(np.vstack([distm[merge_a], distm[merge_b]]),
                                     axis=0)
                distm[:, merge_a] = distm[merge_a]
                # remove columns
                distm = np.delete(distm, merge_b, axis=0)
                distm = np.delete(distm, merge_b, axis=1)
                # fix the diagonal entries
                for idx in range(len(distm)):
                    distm[idx, idx] = np.inf
            else:
                # otherwise find principal directions
                with warnings.catch_warnings():
                    if suppress_external_warnings():
                        warnings.simplefilter('ignore')
                    pca_a = PCA(n_components=1).fit(X_border[clusters[merge_a]])
                    pca_b = PCA(n_components=1).fit(X_border[clusters[merge_b]])
                # extract the angle of principal directions
                numerator = np.dot(pca_a.components_[0], pca_b.components_[0])
                denominator = np.linalg.norm(pca_a.components_[0])
                denominator *= np.linalg.norm(pca_b.components_[0])
                angle = abs(numerator/denominator)
                # check if angle if angle is above a specific threshold
                if angle > self.thres:
                    # do the merge
                    clusters[merge_a] = np.hstack([clusters[merge_a],
                                                   clusters[merge_b]])
                    del clusters[merge_b]
                    # update the distance matrix acoordingly
                    distm[merge_a] = np.min(np.vstack([distm[merge_a], distm[merge_b]]),
                                         axis=0)
                    distm[:, merge_a] = distm[merge_a]
                    # remove columns
                    distm = np.delete(distm, merge_b, axis=0)
                    distm = np.delete(distm, merge_b, axis=1)
                    # fixing the digaonal entries
                    for idx in range(len(distm)):
                        distm[idx, idx] = np.inf
                else:
                    # changing the distance of clusters to fininte
                    distm[merge_a, merge_b] = np.inf
                    distm[merge_b, merge_a] = np.inf

        return clusters

    def determine_indices(self, vectors, nn_params):
        """
        Determines the neighborhood structure within the cluster

        Args:
            vectors (np.array): vectors within the cluster
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array: the neighborhood structure
        """
        n_neighbors = np.min([self.n_neighbors + 1, len(vectors)])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))

        indices = nearestn.fit(vectors).kneighbors(vectors,
                                                    return_distance=False)

        return indices

    def generate_samples_in_clusters(self, vectors, n_to_sample, nn_params):
        """
        Generate samples within the clusters.

        Args:
            vectors (list(np.array)): the vectors of the clusters
            n_to_sample (int): the overall number of samples to generate
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array: the generated samples
        """
        # extract cluster sizes and calculating point distribution in clusters
        # the last element of the clusters is the set of non-border xamples
        cluster_sizes = np.array([len(vect) for vect in vectors])
        cluster_density = cluster_sizes/np.sum(cluster_sizes)

        cluster_indices = self.random_state.choice(len(cluster_sizes),
                                                    n_to_sample,
                                                    p=cluster_density)

        cluster_unique, cluster_count = np.unique(cluster_indices,
                                                    return_counts=True)

        samples = []
        for idx, cluster in enumerate(cluster_unique):
            indices = self.determine_indices(vectors[cluster], nn_params)

            #if len(vectors[cluster]) >= self.n_dim:
            samples.append(self.sample_simplex(X=vectors[cluster],
                                            indices=indices,
                                            n_to_sample=cluster_count[idx]))
            #else:
            #    sample_indices = self.random_state.choice(np.arange(len(vectors[cluster])),
            #                                                cluster_count[idx])
            #    samples.append(vectors[cluster][sample_indices])

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                        self.metric_tensor_from_nn_params(nn_params, X, y)

        X_min = X[y == self.min_label]

        X_border, X_non_border = self.determine_border_non_border(X, y, # pylint: disable=invalid-name
                                                                    X_min,
                                                                    nn_params)

        if len(X_border) == 0:
            return self.return_copies(X, y, "X_border is empty")

        clusters = self.do_the_clustering(X_border, nn_params)

        # extract vectors belonging to the various clusters
        vectors = [X_border[c] for c in clusters if len(c) > 0]
        # adding non-border samples
        if len(X_non_border) > 0:
            vectors.append(X_non_border)

        samples = self.generate_samples_in_clusters(vectors, n_to_sample, nn_params)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'pop': self.pop,
                'thres': self.thres,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the AND_SMOTE method.
"""

import numpy as np

from ..base import coalesce_dict, coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['AND_SMOTE']

class AND_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @inproceedings{and_smote,
                             author = {Yun, Jaesub and Ha,
                                 Jihyun and Lee, Jong-Seok},
                             title = {Automatic Determination of Neighborhood
                                        Size in SMOTE},
                             booktitle = {Proceedings of the 10th International
                                            Conference on Ubiquitous
                                            Information Management and
                                            Communication},
                             series = {IMCOM '16},
                             year = {2016},
                             isbn = {978-1-4503-4142-4},
                             location = {Danang, Viet Nam},
                             pages = {100:1--100:8},
                             articleno = {100},
                             numpages = {8},
                             doi = {10.1145/2857546.2857648},
                             acmid = {2857648},
                             publisher = {ACM},
                             address = {New York, NY, USA},
                             keywords = {SMOTE, imbalanced learning, synthetic
                                            data generation},
                            }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 K=15,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            K (int): maximum number of nearest neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): the simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(K, "K", 2)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.K = K # pylint: disable=invalid-name
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """

        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'K': [9, 15, 21]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def calculate_r_min_maj(self, *, min_idx, neighbor_indices, neigh_idx, X, y):
        """
        Calculate the regions.

        Args:
            min_idx (int): minority index
            neighbor_indices (np.array): the neighbor indices
            neigh_idx (int): one neighbor index
            X (np.array): all features
            y (np.array): all targets

        Returns:
            set, set: the minority and majority regions
        """
        # region coordinates
        reg = np.hstack([min_idx, neighbor_indices[neigh_idx]])

        # all the points in the region must be among the neighbors
        # what we do is counting how many of them are minority and
        # majority samples
        reg_indices= neighbor_indices[:(neigh_idx+1)]
        X_all = X[neighbor_indices[:(neigh_idx+1)]]

        # comparison to the corener points, the mask determines whether
        # the elements of X_all are within the region
        mask = np.logical_and(np.all(np.min(X[reg], axis=0) <= X_all, axis=1),
                             np.all(X_all <= np.max(X[reg], axis=0), axis=1))

        min_mask = np.logical_and(mask, y[reg_indices] == self.min_label)
        maj_mask = np.logical_and(mask, y[reg_indices] == self.maj_label)

        r_min = set(reg_indices[min_mask].tolist())
        r_maj = set(reg_indices[maj_mask].tolist())

        return r_min, r_maj

    def calculate_kappa(self, min_idx, neighbor_indices, X, y):
        """
        Calculates the kappa measure as specified in the paper.

        Args:
            min_idx (int): the index of a minority sample
            neighbor_indices (np.array): the indices of its neighbors
            X (np.array): all features
            y (np.array): all targets

        Returns:
            int: the kappa measure
        """
        regions_min = []
        regions_maj = []

        for neigh_idx in range(1, neighbor_indices.shape[0]):
            # continuing if the label of the neighbors is not minority
            if y[neighbor_indices[neigh_idx]] == self.maj_label:
                continue

            r_min, r_maj = self.calculate_r_min_maj(min_idx=min_idx,
                                            neighbor_indices=neighbor_indices,
                                            neigh_idx=neigh_idx,
                                            X=X,
                                            y=y)

            # appending the coordinates of points to the minority and
            # majority regions
            if len(regions_min) == 0:
                regions_min.append(r_min)
                regions_maj.append(r_maj)
            else:
                regions_min.append(regions_min[-1].union(r_min))
                regions_maj.append(regions_maj[-1].union(r_maj))

        # computing the lengths of the increasing minority and majority
        # sets
        regions_min = np.array([len(r) for r in regions_min])
        regions_maj = np.array([len(r) for r in regions_maj])

        # computing the precision of minority classification (all points
        # are supposed to be classified as minority)
        prec = regions_min/(regions_min + regions_maj)
        # discrete differentiation with order 1
        discrete_difference = np.diff(prec, 1)
        # finding the biggest drop (+1 because diff reduces length, +1
        # because of indexing begins with 0)
        if len(discrete_difference) == 0:
            return 0

        # returning the coordinate of the biggest drop as the ideal
        # neighborhood size note that k indices the minority neighbors
        return np.argmin(discrete_difference) + 2

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        K = min([len(X_min), self.K]) # pylint: disable=invalid-name

        # find K nearest neighbors of all samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=K,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X)
        ind = nearestn.kneighbors(X, return_distance=False)

        # indices of minority samples
        min_ind = np.where(y == self.min_label)[0]

        # Executing the algorithm
        # iteration through all minority samples
        # kappa stores the ideal neighborhood sizes for all minority samples
        kappa = []
        for _, min_idx in enumerate(min_ind):
            kappa.append(self.calculate_kappa(min_idx, ind[min_idx], X, y))

        # finding nearest minority neighbors of minority samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=max(kappa) + 1,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X_min)
        ind = nearestn.kneighbors(X_min, return_distance=False)

        if np.sum(kappa) == 0:
            return self.return_copies(X, y, "No minority samples in nearest neighbors")

        # turning the neighbor adjecency matrix into a weight matrix
        # based on the kappa scores

        weights = ind.copy()
        weights[:,:]= 1.0
        for idx in range(weights.shape[0]):
            if kappa[idx] == 0:
                weights[idx, :]= 0.0
            else:
                weights[idx, (kappa[idx]+1):]= 0.0

        samples = self.sample_simplex(X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        simplex_weights=weights)

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'K': self.K,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the SPY method.
"""

import numpy as np

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['SPY']

class SPY(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{spy,
                            author={Dang, X. T. and Tran, D. H. and Hirose, O.
                                    and Satou, K.},
                            booktitle={2015 Seventh International Conference
                                        on Knowledge and Systems Engineering
                                        (KSE)},
                            title={SPY: A Novel Resampling Method for
                                    Improving Classification Performance in
                                    Imbalanced Data},
                            year={2015},
                            volume={},
                            number={},
                            pages={280-285},
                            keywords={decision making;learning (artificial
                                        intelligence);pattern classification;
                                        sampling methods;SPY;resampling
                                        method;decision-making process;
                                        biomedical data classification;
                                        class imbalance learning method;
                                        SMOTE;oversampling method;UCI
                                        machine learning repository;G-mean
                                        value;borderline-SMOTE;
                                        safe-level-SMOTE;Support vector
                                        machines;Training;Bioinformatics;
                                        Proteins;Protein engineering;Radio
                                        frequency;Sensitivity;Imbalanced
                                        dataset;Over-sampling;
                                        Under-sampling;SMOTE;
                                        borderline-SMOTE},
                            doi={10.1109/KSE.2015.24},
                            ISSN={},
                            month={Oct}}
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 threshold=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_neighbors (int): number of neighbors in nearest neighbor
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            threshold (float): threshold*n_neighbors gives the threshold z
                                described in the paper
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_in_range(threshold, "threshold", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.threshold = threshold
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'n_neighbors': [3, 5, 7],
                                  'threshold': [0.3, 0.5, 0.7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_min = X[y == self.min_label]

        # fitting nearest neighbors model
        n_neighbors = min([len(X), self.n_neighbors + 1])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                        self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        # if the number of majority samples in the neighborhood is
        # smaller than a threshold
        # their labels are changed to minority

        y_new = y.copy()
        z_vec = self.threshold * n_neighbors

        majority_mask = y[ind[:, 1:]] == self.maj_label

        x_vec = np.sum(majority_mask, axis=1)
        mask = x_vec < z_vec

        mask = list(set(ind[mask, 1:][majority_mask[mask]].tolist()))

        y_new[mask] = self.min_label

        # checking the neighbors of each minority sample
        #for i in range(len(X_min)):
        #    majority_mask = y[ind[i][1:]] == self.maj_label
        #    x = np.sum(majority_mask)
        #    # if the number of majority samples in the neighborhood is
        #    # smaller than a threshold
        #    # their labels are changed to minority
        #    if x < z:
        #        y_new[ind[i][1:][majority_mask]] = self.min_label

        return X.copy(), y_new

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'threshold': self.threshold,
                **OverSampling.get_params(self)}



"""
This module implements the RWO_sampling method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['RWO_sampling']

class RWO_sampling(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{rwo_sampling,
                    author = {Zhang, Huaxzhang and Li, Mingfang},
                    year = {2014},
                    month = {11},
                    pages = {},
                    title = {RWO-Sampling: A Random Walk Over-Sampling Approach
                                to Imbalanced Data Classification},
                    volume = {20},
                    booktitle = {Information Fusion}
                    }
    """

    categories = [OverSamplingSimplex.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 1, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        stds = np.std(X_min, axis=0)/np.sqrt(len(X_min))

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # fitting nearest neighbors model to all samples
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=self.n_dim,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nearestn.fit(X_min)
        indices = nearestn.kneighbors(X_min, return_distance=False)

        samples = self.sample_simplex(X=X_min,
                                        indices=indices,
                                        n_to_sample=n_to_sample)

        samples = samples \
                + self.random_state.normal(size=samples.shape) * stds

        return (np.vstack([X, samples]),
                np.hstack([y, np.array([self.min_label]*len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the SMOTE_TomekLinks method.
"""

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ..noise_removal import TomekLinkRemoval
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['SMOTE_TomekLinks']

class SMOTE_TomekLinks(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_tomeklinks_enn,
                     author = {Batista, Gustavo E. A. P. A. and Prati,
                                Ronaldo C. and Monard, Maria Carolina},
                     title = {A Study of the Behavior of Several Methods for
                                Balancing Machine Learning Training Data},
                     journal = {SIGKDD Explor. Newsl.},
                     issue_date = {June 2004},
                     volume = {6},
                     number = {1},
                     month = jun,
                     year = {2004},
                     issn = {1931-0145},
                     pages = {20--29},
                     numpages = {10},
                     url = {http://doi.acm.org/10.1145/1007730.1007735},
                     doi = {10.1145/1007730.1007735},
                     acmid = {1007735},
                     publisher = {ACM},
                     address = {New York, NY, USA},
                    }
    """

    categories = [OverSampling.cat_sample_ordinary,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the SMOTE object

        Args:
            proportion (float): proportion of the difference of n_maj and
                                n_min to sample e.g. 1.0 means that after
                                sampling the number of minority samples
                                will be equal to the number of majority
                                samples
            n_neighbors (int): control parameter of the nearest neighbor
                                technique
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            preferential (bool): whether to use preferential connectivity in the
                                    neighborhoods
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        return SMOTE.parameter_combinations(raw)

    def return_results(self, X_samp, y_samp, X, y):
        """
        Prepare the final dataset.

        Args:
            X_samp (np.array): the sampled vectors
            y_samp (np.array): the labels of the sampled vectors
            X (np.array): the original training vectors
            y (np.array): the original target labels

        Returns:
            np.array, np.array: the oversampled dataset and the labels
        """
        if len(X_samp) == 0:
            return self.return_copies(X, y, "All samples have been removed, "\
                                            "returning the original dataset.")
        return X_samp, y_samp

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        smote = SMOTE(self.proportion,
                      self.n_neighbors,
                      nn_params=nn_params,
                      ss_params=self.ss_params,
                      n_jobs=self.n_jobs,
                      random_state=OverSampling.get_params(self)['random_state'])
        X_new, y_new = smote.sample(X, y)

        tomek = TomekLinkRemoval(strategy='remove_both',
                                n_jobs=self.n_jobs,
                                nn_params=nn_params)

        X_samp, y_samp = tomek.remove_noise(X_new, y_new)

        return self.return_results(X_samp, y_samp, X, y)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the LN_SMOTE method.
"""

import numpy as np

from ..base import coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['LN_SMOTE']

class LN_SMOTE(OverSampling): # pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{ln_smote,
                            author={Maciejewski, T. and Stefanowski, J.},
                            booktitle={2011 IEEE Symposium on Computational
                                        Intelligence and Data Mining (CIDM)},
                            title={Local neighbourhood extension of SMOTE for
                                        mining imbalanced data},
                            year={2011},
                            volume={},
                            number={},
                            pages={104-111},
                            keywords={Bayes methods;data mining;pattern
                                        classification;local neighbourhood
                                        extension;imbalanced data mining;
                                        focused resampling technique;SMOTE
                                        over-sampling method;naive Bayes
                                        classifiers;Noise measurement;Noise;
                                        Decision trees;Breast cancer;
                                        Sensitivity;Data mining;Training},
                            doi={10.1109/CIDM.2011.5949434},
                            ISSN={},
                            month={April}}
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_componentwise,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): parameter of the NearestNeighbors component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0.0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def safe_level(self, p_idx, *, y, indices, n_idx=None):
        """
        computing the safe level of samples

        Args:
            p_idx (int): index of positive sample
            y (np.array): the target labels
            indices (np.array): the neighborhood structure
            n_idx (int): index of other sample

        Returns:
            int: safe level
        """
        if n_idx is None:
            # implementation for 1 sample only
            return np.sum(y[indices[p_idx][1:-1]] == self.min_label)

        # implementation for 2 samples
        if ((not y[n_idx] != self.maj_label)
                and p_idx in indices[n_idx][1:-1]):
            # -1 because p_idx will be replaced
            n_positives = np.sum(y[indices[n_idx][1:-1]] == self.min_label) - 1
            if y[indices[n_idx][-1]] == self.min_label:
                # this is the effect of replacing p_idx by the next
                # (k+1)th neighbor
                n_positives = n_positives + 1
            return n_positives
        return np.sum(y[indices[n_idx][1:-1]] == self.min_label)

    def random_gap(self, *, slp, sln, n_label, n_neighbors, n_dim):
        """
        determining random gap

        Args:
            slp (int): safe level of p
            sln (int): safe level of n
            n_label (int): label of n
            n_neighbors (int): number of neighbors
            n_dim (int): number of dimensions

        Returns:
            float: gap
        """
        delta = 0.0

        if sln == 0 and slp > 0:
            return np.repeat(delta, n_dim)

        sl_ratio = slp/sln

        if sl_ratio == 1:
            delta = self.random_state.random_sample(size=n_dim)
        elif sl_ratio > 1:
            delta = self.random_state.random_sample(size=n_dim)/sl_ratio
        else:
            delta = 1.0 - self.random_state.random_sample(size=n_dim)*sl_ratio

        if not n_label == self.min_label:
            delta = delta * sln / n_neighbors

        return delta

    def calculate_slp_sln(self, minority_indices, indices, y):
        """
        Calculate the safe levels

        Args:
            minority_indices (np.array): the minority indices
            indices (np.array): the neighborhood structure
            y (np.array): the target labels

        Returns:
            dict: safe levels of positives and their
                                neighbors
        """
        slpsln = {}

        for p_idx in minority_indices:
            slp = self.safe_level(p_idx, y=y, indices=indices)
            for n_idx in indices[p_idx][1:-1]:
                sln = self.safe_level(p_idx, n_idx=n_idx, y=y,
                                            indices=indices)
                slpsln[(p_idx, n_idx)] = (slp, sln)

        return slpsln

    def generate_samples(self, *, slpsln, n_to_sample, X, y, n_neighbors):
        """
        Generate samples

        Args:
            slpsln (dict): the safe level values
            n_to_sample (int): number of samples to generate
            X (np.array): all feature vectors
            y (np.array): all target labels
            n_neighbors (int): the number of neighbors to use

        Returns:
            np.array: the generated samples
        """

        if len(slpsln) == 0:
            return np.zeros(shape=(0, X.shape[1]))

        samples = []
        # generating samples
        while len(samples) < n_to_sample:
            idx = self.random_state.choice(np.arange(len(slpsln)))
            p_idx, n_idx = list(slpsln.keys())[idx]

            slp, sln = slpsln[(p_idx, n_idx)]

            delta = self.random_gap(slp=slp,
                                    sln=sln,
                                    n_label=y[n_idx],
                                    n_neighbors=n_neighbors,
                                    n_dim=X.shape[1])

            samples.append(X[p_idx] + (X[n_idx] - X[p_idx]) * delta)

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        n_neighbors = np.min([len(X) - 2, self.n_neighbors])

        if n_neighbors < 2:
            return self.return_copies(X, y, f"Too few samples {len(X)}")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # nearest neighbors of each instance to each instance in the dataset
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors + 2,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices = nnmt.kneighbors(X, return_distance=False)

        minority_indices = np.where(y == self.min_label)[0]

        # calculate safe levels
        slpsln = self.calculate_slp_sln(minority_indices, indices, y)

        slpsln = {key: item for key, item in slpsln.items()\
                                            if item[0] > 0 and item[1] > 0}

        samples = self.generate_samples(slpsln=slpsln,
                                        n_to_sample=n_to_sample,
                                        X=X,
                                        y=y,
                                        n_neighbors=n_neighbors)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements an sklearn compatible classifier
with oversampling.
"""
from sklearn.base import BaseEstimator, ClassifierMixin

from smote_variants.base import instantiate_obj

class OversamplingClassifier(BaseEstimator, ClassifierMixin):
    """
    This class wraps an oversampler and a classifier, making it compatible
    with sklearn based pipelines.
    """

    def __init__(self, oversampler, classifier):
        """
        Constructor of the wrapper.

        Args:
            oversampler (obj): an oversampler object
            classifier (obj): an sklearn-compatible classifier
        """

        self.oversampler = oversampler
        self.classifier = classifier

        self.oversampler_obj = instantiate_obj(oversampler)
        self.classifier_obj = instantiate_obj(classifier)

    def fit(self, X, y=None):
        """
        Carries out oversampling and fits the classifier.

        Args:
            X (np.ndarray): feature vectors
            y (np.array): target values

        Returns:
            obj: the object itself
        """

        X_samp, y_samp = self.oversampler_obj.sample(X, y)
        self.classifier_obj.fit(X_samp, y_samp)

        return self

    def predict(self, X):
        """
        Carries out the predictions.

        Args:
            X (np.ndarray): feature vectors
        """

        return self.classifier_obj.predict(X)

    def predict_proba(self, X):
        """
        Carries out the predictions with probability estimations.

        Args:
            X (np.ndarray): feature vectors
        """

        return self.classifier_obj.predict_proba(X)

    def get_params(self, deep=True):
        """
        Returns the dictionary of parameters.

        Args:
            deep (bool): wether to return parameters with deep discovery

        Returns:
            dict: the dictionary of parameters
        """

        return {'oversampler': self.oversampler,
                'classifier': self.classifier}

    def set_params(self, **parameters):
        """
        Sets the parameters.

        Args:
            parameters (dict): the parameters to set.

        Returns:
            obj: the object itself
        """

        for parameter, value in parameters.items():
            setattr(self, parameter, value)

        self.oversampler_obj = instantiate_obj(self.oversampler)
        self.classifier_obj = instantiate_obj(self.classifier)

        return self



"""
Testing the TomekLinkRemoval.
"""

from smote_variants.noise_removal import TomekLinkRemoval

from smote_variants.datasets import load_all_min_noise

def test_specific():
    """
    Oversampler specific testing
    """
    dataset = load_all_min_noise()

    obj = TomekLinkRemoval(strategy='remove_both')

    X_samp, _ = obj.remove_noise(dataset['data'], dataset['target'])

    assert len(X_samp) > 0



"""
This module implements the NoSMOTE method.
"""

from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['NoSMOTE']

class NoSMOTE(OverSampling):
    """
    The goal of this class is to provide a functionality to send data through
    on any model selection/evaluation pipeline with no oversampling carried
    out. It can be used to get baseline estimates on preformance.
    """

    categories = []

    def __init__(self, raise_value_error=False,
                        raise_runtime_error=False,
                        random_state=None,
                        **_kwargs):
        """
        Constructor of the NoSMOTE object.

        Args:
            random_state (int/np.random.RandomState/None): dummy parameter for \
                        the compatibility of interfaces
            raise_value_error (bool): whether to raise a ValueError
            raise_runtime_error (bool): whether to raise a RuntimeError
        """
        super().__init__(random_state=random_state)

        self.raise_value_error = raise_value_error
        self.raise_runtime_error = raise_runtime_error

    @classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            raw (bool): whether to return the raw parameter generation setup
        """
        return cls.generate_parameter_combinations({}, raw=raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        if self.raise_value_error:
            raise ValueError("Dummy ValueError raised by NoSMOTE")

        if self.raise_runtime_error:
            raise RuntimeError("Dummy RuntimeError raised by NoSMOTE")

        return X.copy(), y.copy()

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'raise_value_error': self.raise_value_error,
                'raise_runtime_error': self.raise_runtime_error,
                **OverSampling.get_params(self)}



"""
This module implements the ADG method.
"""

import warnings

import numpy as np
from numpy.linalg import LinAlgError

from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.metrics import pairwise_distances

from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['ADG']

def bic_score(kmeans, X):
    """
    Compute BIC score for clustering

    Args:
        kmeans (sklearn.KMeans): kmeans object
        X (np.array):  clustered data

    Returns:
        float: bic value

    Inspired by https://stats.stackexchange.com/questions/90769/
                        using-bic-to-estimate-the-number-of-k-in-kmeans
    """

    # extract descriptors of the clustering
    cluster_centers = kmeans.cluster_centers_
    #n_clusters = kmeans.n_clusters
    n_clusters = np.unique(kmeans.labels_).shape[0]
    n_in_clusters = np.bincount(kmeans.labels_)
    n_samples, n_dim = X.shape

    # compute variance for all clusters beforehand

    def sum_norm_2(idx):
        return np.sum(np.linalg.norm(X[kmeans.labels_ == idx] -
                                        cluster_centers[idx])**2)

    cluster_variances = [sum_norm_2(i) for i in range(n_clusters)]
    term_0 = (1.0)/((n_samples - n_clusters) * n_dim)
    term_1 = np.sum(cluster_variances)
    clustering_variance = term_0 * term_1

    const_term = 0.5 * n_clusters * np.log(n_samples) * (n_dim+1)

    def bic_comp(idx):
        term_0 = n_in_clusters[idx] * np.log(n_in_clusters[idx])
        term_1 = n_in_clusters[idx] * np.log(n_samples)
        if clustering_variance == 0:
            return np.inf
        term_2 = (((n_in_clusters[idx] * n_dim) / 2)
                    * np.log(2*np.pi*clustering_variance))
        term_3 = ((n_in_clusters[idx] - 1) * n_dim / 2)

        return term_0 - term_1 - term_2 - term_3

    bic = np.sum([bic_comp(idx) for idx in range(n_clusters)
                                if n_in_clusters[idx] > 0]) - const_term

    return bic

def xmeans(X, rng=(1, 10), random_state=None):
    """
    Clustering with BIC based n_cluster selection

    Args:
        X (np.array): data to cluster
        rng (tuple): lower and upper bound on the number of clusters
        random_state (int/None): random state to use

    Returns:
        sklearn.KMeans: clustering with lowest BIC score
    """
    best_bic = np.inf
    best_clustering = None

    # do clustering for all n_clusters in the specified range
    for n_clusters in range(rng[0], min([rng[1], len(X)])):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            kmeans = KMeans(n_clusters=n_clusters,
                            random_state=random_state).fit(X)

        bic = bic_score(kmeans, X)
        if bic < best_bic:
            best_bic = bic
            best_clustering = kmeans

    return best_clustering

def xgmeans(X, rng=(1, 10), random_state=None):
    """
    Gaussian mixture with BIC to select the optimal number
    of components

    Args:
        X (np.array): data to cluster
        rng (tuple): lower and upper bound on the number of components
        random_state (int/None): random state to use

    Returns:
        sklearn.GaussianMixture: Gaussian mixture model with the
                                    lowest BIC score
    """
    best_bic = np.inf
    best_mixture = None

    # do model fitting for all n_components in the specified range
    for n_components in range(rng[0], min([rng[1], len(X)])):
        gmm = GaussianMixture(n_components=n_components,
                                random_state=random_state).fit(X)
        bic = gmm.bic(X)
        if bic < best_bic:
            best_bic = bic
            best_mixture = gmm

    return best_mixture

def evaluate_matrices(X, y, kernel=np.inner, maj_label=0, min_label=1):
    """
    The function evaluates the partial_results specified in the method.

    Args:
        X (np.array): features
        y (np.array): target labels
        kernel (function): the kernel function to be used

    Returns:
        np.array, np.array, int, int, np.array, np.array,
        np.array, np.array, np.array
        np.array, np.array, np.array, np.array, np.array:
            X_minux, X_plus, l_minus, l_plus, X, y, K, M_plus, M_minus,
            M, K_plus, K_minus, N_plus, n_minus using the notations of
            the paper, X and y are ordered by target labels
    """
    x_minus = X[y == maj_label]
    x_plus = X[y == min_label]

    X = np.vstack([x_minus, x_plus])
    y = np.hstack([np.array([maj_label]*len(x_minus)),
                    np.array([min_label]*len(x_plus))])

    k = pairwise_distances(X, X, metric=kernel) # pylint: disable=invalid-name
    m_plus = np.mean(k[:, len(x_minus):], axis=1)
    m_minus = np.mean(k[:, :len(x_minus)], axis=1)
    m = np.dot(m_minus - m_plus, m_minus - m_plus) # pylint: disable=invalid-name

    k_minus = k[:, :len(x_minus)]
    k_plus = k[:, len(x_minus):]

    partial_results = {'X_minus': x_minus,
                'X_plus': x_plus,
                'K': k,
                'M_plus': m_plus,
                'M_minus': m_minus,
                'M': m,
                'K_plus': k_plus,
                'K_minus': k_minus,
                'l_minus': len(x_minus),
                'l_plus': len(x_plus)}

    return (partial_results, X, y)


def generate_kernel_function(specification):
    """
    Generate the kernel function.

    Args:
        specification (str): the kernel specification

    Returns:
        callable: the kernel function
    """
    if specification == 'inner':
        kernel_function = np.inner
    else:
        kernel_function = None
        specs = specification.split('_')
        if specs[0] == 'rbf':
            bandwidth = float(specs[1])
            def kernel_function(x_vector, y_vector):
                norm = np.linalg.norm(x_vector - y_vector)
                return np.exp(-norm**2/bandwidth)

    return kernel_function

class ADGSingularMatrixException(Exception):
    """
    Exception indicating a singular matrix problem.
    """

class ADG(OverSampling):
    """
    References:
        * BibTex::

            @article{adg,
                    author = {Pourhabib, A. and Mallick, Bani K. and Ding, Yu},
                    year = {2015},
                    month = {16},
                    pages = {2695--2724},
                    title = {A Novel Minority Cloning Technique for
                                Cost-Sensitive Learning},
                    volume = {16},
                    journal = {Journal of Machine Learning Research}
                    }

    Notes:
        * This method has a lot of parameters, it becomes fairly hard to
            cross-validate thoroughly.
        * Fails if matrix is singular when computing alpha_star, fixed
            by PCA.
        * Singularity might be caused by repeating samples.
        * Maintaining the kernel matrix becomes unfeasible above a couple
            of thousand vectors.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_uses_clustering]

    def __init__(self,
                 proportion=1.0,
                 *,
                 kernel='inner',
                 lam=1.0,
                 mu=1.0,
                 k=12,
                 gamma=1.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            kernel (str): 'inner'/'rbf_x', where x is a float, the bandwidth
            lam (float): lambda parameter of the method
            mu (float): mu parameter of the method
            k (int): number of samples to generate in each iteration
            gamma (float): gamma parameter of the method
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state, checks=None)

        self.check_greater_or_equal(proportion, "proportion", 0)

        if kernel != 'inner' and not kernel.startswith('rbf'):
            raise ValueError(f"{self.__class__.__name__}: Kernel function "\
                                f"{kernel} not supported")
        if kernel.startswith('rbf'):
            par = float(kernel.split('_')[-1])
            if par <= 0.0:
                raise ValueError(f"{self.__class__.__name__}: Kernel parameter "\
                                    f"{par} is not supported")

        self.check_greater(lam, 'lam', 0)
        self.check_greater(mu, 'mu', 0)
        self.check_greater_or_equal(k, 'k', 1)
        self.check_greater(gamma, 'gamma', 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.kernel = kernel
        self.lam = lam
        self.mu = mu # pylint: disable=invalid-name
        self.k = k # pylint: disable=invalid-name
        self.gamma = gamma
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'kernel': ['inner', 'rbf_0.5',
                                             'rbf_1.0', 'rbf_2.0'],
                                  'lam': [1.0, 2.0],
                                  'mu': [1.0, 2.0],
                                  'k': [12],
                                  'gamma': [1.0, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def initialize_matrices(self, *, partial_results):
        """
        Initializes the base partial_results.

        Args:
            np.arrays by names

        Returns:
            np.arrays: the initialized partial_results
        """
        l_plus = partial_results['l_plus']
        k_plus = partial_results['K_plus']
        k_minus = partial_results['K_minus']
        x_plus = partial_results['X_plus']
        x_minus = partial_results['X_minus']

        k_plus2 = np.dot(k_plus, k_plus.T)
        k_plus_sum = np.sum(k_plus, axis=1)
        k_plus_diad = np.outer(k_plus_sum, k_plus_sum)/l_plus

        k_minus2 = np.dot(k_minus, k_minus.T)
        k_minus_sum = np.sum(k_minus, axis=1)
        k_minus_diad = np.outer(k_minus_sum,
                                k_minus_sum)/partial_results['l_minus']

        n = k_plus2 - k_plus_diad + k_minus2 - k_minus_diad # pylint: disable=invalid-name

        x_plus_hat = x_plus.copy()

        partial_results['K_plus2'] = k_plus2
        partial_results['K_plus_sum'] = k_plus_sum
        partial_results['K_minus2'] = k_minus2
        partial_results['K_minus_sum'] = k_minus_sum
        partial_results['N'] = n
        partial_results['X_plus_hat'] = x_plus_hat
        partial_results['l_minus'] = len(x_minus)

        return partial_results

    def sample_with_n_components(self, X, y, n_components):
        """
        Rerun the sampling in a lower dimensional space.

        Args:
            X (np.array): features
            y (np.array): targets
            n_components (int): dimensionality of the feature space

        Returns:
            np.array, np.array: the resampled X and y
        """
        _logger.warning("%s: reducing dimensionality to %d",
                        self.__class__.__name__, n_components)
        #if n_components == 0:
        #    return None, None

        pca = PCA(n_components=n_components).fit(X)

        x_trans = pca.transform(X)

        adg = ADG(proportion=self.proportion,
                    kernel=self.kernel,
                    lam=self.lam,
                    mu=self.mu,
                    k=self.k,
                    gamma=self.gamma,
                    random_state=self._random_state_init)

        X_samp, y_samp = adg.sample(x_trans, y)

        return pca.inverse_transform(X_samp), y_samp

    def update_partial_results(self, *, partial_results):
        """
        Update partial_results.

        Args:
            np.arrays: the constants and partial_results used in the paper

        Returns:
            np.array: the updated partial_results
        """
        # M_plus is not updated as it is not used

        k_plus = partial_results['K_plus']
        k_minus = partial_results['K_minus']
        k_minus2 = partial_results['K_minus2']
        k_plus2 = partial_results['K_plus2']
        l_new = partial_results['l_new']

        m_plus = np.mean(k_plus, axis=1)
        m_minus = np.mean(k_minus, axis=1)

        # step 13 is already involved in the core of the loop
        m = np.dot(m_minus - m_plus, m_minus - m_plus) # pylint: disable=invalid-name

        corner = np.dot(k_minus[:-l_new:], k_minus[-l_new:].T)
        k_minus2 = np.block([[k_minus2, corner],
                                [corner.T, np.dot(k_minus[-l_new:],
                                                    k_minus[-l_new:].T)]])
        k_minus_sum = m_minus*len(k_minus)

        k_plus2 = k_plus2 + np.dot(k_plus[:-l_new, l_new:],
                                    k_plus[:-l_new, l_new:].T)

        corner = np.dot(k_plus[:-l_new], k_plus[-l_new:].T)

        k_plus2 = np.block([[k_plus2, corner],
                            [corner.T, np.dot(k_plus[-l_new:],
                                                    k_plus[-l_new:].T)]])

        k_plus_sum = m_plus*len(k_plus)

        n = k_plus2 - np.outer(k_plus_sum/partial_results['l_plus'], # pylint: disable=invalid-name
                                k_plus_sum) + \
            k_minus2 - np.outer(k_minus_sum/partial_results['l_minus'],
                                k_minus_sum)

        partial_results['M_plus'] = m_plus
        partial_results['M_minus'] = m_minus
        partial_results['M'] = m
        partial_results['K_minus2'] = k_minus2
        partial_results['K_minus_sum'] = k_minus_sum
        partial_results['K_plus2'] = k_plus2
        partial_results['K_plus_sum'] = k_plus_sum
        partial_results['N'] = n

        return partial_results


    def determine_a(self, *, partial_results, k_c, lam_c, mu_c):
        """
        Determine the matrix A

        Args:
            the quantities required for A

        Returns:
            np.array: the matrix A
        """
        m = partial_results['M'] # pylint: disable=invalid-name
        n = partial_results['N'] # pylint: disable=invalid-name
        k = partial_results['K'] # pylint: disable=invalid-name

        # step 3, 4
        omega = - np.sum([k_c[i]*(lam_c[i])**2/(4*mu_c[i]**2)
                            for i in range(len(k_c))])
        A = (m - self.gamma*n) - omega*k # pylint: disable=invalid-name

        return A

    def determine_b(self, *, partial_results, k_c, lam_c, X, clusters):
        """
        Determine the vector b

        Args:
            the quantities required for vector b

        Returns:
            np.array: the vector b
        """
        x_minus = partial_results['X_minus']
        k = partial_results['K'] # pylint: disable=invalid-name
        m_minus = partial_results['M_minus']

        # step 3, 4
        nu_c = - 0.5*k_c*lam_c
        m_plus_c = [np.mean(k[:, np.arange(len(x_minus), len(X))[
            clusters.labels_ == i]]) for i in range(len(k_c))]

        b = np.sum([(m_minus - m_plus_c[i])*nu_c[i] # pylint: disable=invalid-name
                    for i in range(len(k_c))], axis=0)

        return b

    def step_1_2_3_4(self, *, partial_results, X):
        """
        Implements steps 1, 2, 3 and 4 of the paper.

        Args:
            np.arrays: the constants and partial_results used in the paper

        Returns:
            np.array: the updated partial_results
        """
        x_plus_hat = partial_results['X_plus_hat']

        # step 1
        clusters = xmeans(x_plus_hat, random_state = self._random_state_init)
        l_c = np.array([np.sum(clusters.labels_ == i)
                        for i in range(clusters.n_clusters)])

        # step 2
        k_c = ((1.0/l_c)/(np.sum(1.0/l_c))*self.k).astype(int)
        k_c[k_c == 0] = 1
        lam_c, mu_c = self.lam/l_c, self.mu/l_c

        return (self.determine_a(partial_results=partial_results, k_c=k_c, lam_c=lam_c,
                                    mu_c=mu_c),
                self.determine_b(partial_results=partial_results, k_c=k_c, lam_c=lam_c,
                                    X=X, clusters=clusters))

    def step_5_6_7(self, *, partial_results, q, X, kernel_function, alpha_star):
        """
        Implements steps 5, 6, 7 of the paper.

        Args:
            np.arrays: the constants and partial_results used in the paper

        Returns:
            np.array: the updated partial_results
        """
        x_plus = partial_results['X_plus']

        # step 5
        mixture = xgmeans(x_plus, random_state=self._random_state_init)

        # step 6
        z = mixture.sample(q)[0] # pylint: disable=invalid-name

        # step 7
        # computing the kernel matrix of generated samples with all samples
        k_10 = pairwise_distances(z, X, metric=kernel_function)
        mask_inner_prod = np.where(np.inner(k_10, alpha_star) > 0)[0]
        z_hat = z[mask_inner_prod]

        partial_results['Z_hat'] = z_hat
        partial_results['mask_inner_prod'] = mask_inner_prod
        partial_results['K_10'] = k_10

    def step_9_16(self, *, partial_results, kernel_function, y):
        """
        Implements steps 9-16 of the paper.

        Args:
            np.arrays: the constants and partial_results used in the paper

        Returns:
            np.array: the updated partial_results
        """
        # step 8
        # this step is not used for anything, the identified clusters are
        # only used in step 13 of the paper, however, the values set
        # (M_plus^c) are overwritten in step 3 of the next iteration

        if partial_results['Z_hat'].shape[0] == 0:
            raise ValueError("Z_hat is empty")

        x_minus = partial_results['X_minus']
        x_plus_hat = partial_results['X_plus_hat']
        k = partial_results['K'] # pylint: disable=invalid-name
        k_10 = partial_results['K_10']
        l_minus = partial_results['l_minus']
        z_hat = partial_results['Z_hat']
        mask_inner_prod = partial_results['mask_inner_prod']

        # step 9
        x_plus_hat = np.vstack([x_plus_hat, z_hat])

        # step 11 - 16
        # these steps have been reorganized a bit for efficient
        # calculations

        pairwd = pairwise_distances(z_hat, z_hat, metric=kernel_function)
        k = np.block([[k, k_10[mask_inner_prod].T], # pylint: disable=invalid-name
                        [k_10[mask_inner_prod], pairwd]])

        k_minus = k[:, :l_minus]
        k_plus = k[:, l_minus:]

        # step 10
        X = np.vstack([x_minus, x_plus_hat])
        y = np.hstack([y, np.repeat(self.min_label, len(z_hat))])

        partial_results['X_plus_hat'] = x_plus_hat
        partial_results['K'] = k
        partial_results['K_minus'] = k_minus
        partial_results['K_plus'] = k_plus
        partial_results['l_plus'] = len(x_plus_hat)
        partial_results['l_new'] = len(partial_results['Z_hat'])

        return X, y

    def solve(self,
                    A, # pylint: disable=invalid-name
                    b  # pylint: disable=invalid-name
                    ):
        """
        Solve a linear equation and throw exception if doesnt work.

        Args:
            A (np.array): the coefficients
            b (np.array): the right hand side

        Returns:
            np.array: the solution
        """
        try:
            return np.linalg.solve(A, b)
        except LinAlgError as exc:
            raise ADGSingularMatrixException() from exc

    def check_early_stopping(self, fraction):
        """
        Checking the early stopping condition.
        """
        if fraction < 0.01:
            raise ValueError("Early stopping.")

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # Implementation of the technique, following the steps and notations
        # of the paper
        n_to_sample = self.det_n_to_sample(self.proportion)

        q = n_to_sample # pylint: disable=invalid-name

        if q == 0:
            return self.return_copies(X, y, "no need for sampling")

        # instantiating the proper kernel function, the parameter of the RBF
        # is supposed to be the denominator in the Gaussian
        kernel_function = generate_kernel_function(self.kernel)

        # Initial evaluation of the partial_results
        # M_plus is not updated as it is not used
        # the partial_results dict acts like a dataclass to reduce the clutter
        # of passing around dozens of arrays
        partial_results, X, y = evaluate_matrices(X, y, kernel=kernel_function)

        # The computing of N matrix is factored into two steps, computing
        # N_plus and N_minus this is used to improve efficiency
        partial_results = self.initialize_matrices(partial_results=partial_results)

        total_added = 0

        alpha_star = None

        q = q * 2 # pylint: disable=invalid-name

        try:
            # executing the sample generation
            while q > 1:
                q = int(q/2) # pylint: disable=invalid-name

                _logger.info("%s: Starting iteration with q=%d",
                                self.__class__.__name__ , q)

                A, b = self.step_1_2_3_4(partial_results=partial_results, X=X) # pylint: disable=invalid-name

                alpha_star = self.solve(A, b)

                self.step_5_6_7(partial_results=partial_results,
                                q=q,
                                X=X,
                                kernel_function=kernel_function,
                                alpha_star=alpha_star)

                _logger.info("%s: number of vectors added: %d/%d",
                            self.__class__.__name__, len(partial_results['Z_hat']), q)

                X, y = self.step_9_16(partial_results=partial_results,
                                        kernel_function=kernel_function,
                                        y=y)

                total_added = total_added + len(partial_results['Z_hat'])

                self.check_early_stopping(len(partial_results['Z_hat'])/total_added)

                partial_results = self.update_partial_results(partial_results=partial_results)

        except ADGSingularMatrixException:
            n_components = int(X.shape[1]/2)
            if n_components > 0:
                return self.sample_with_n_components(X, y, n_components)

            return self.return_copies(X, y, "cannot reduce dimensionality any more")
        except ValueError as exc:
            _logger.info("%s: stopping the iteration because of %s",
                        self.__class__.__name__, str(exc))
            return X.copy(), y.copy()

        return X.copy(), y.copy()

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'kernel': self.kernel,
                'lam': self.lam,
                'mu': self.mu,
                'k': self.k,
                'gamma': self.gamma,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the DBSMOTE method.
"""

import numpy as np

from sklearn.preprocessing import StandardScaler
from sklearn.cluster import DBSCAN

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['DBSMOTE']

DB_LIMIT = 5

# Bellman-Ford algorithm, inspired by
# https://gist.github.com/joninvski/701720
def initialize(graph, source):
    """
    Initializes shortest path algorithm.

    Args:
        graph (dict): graph in dictionary representation
        source (key): source node

    Returns:
        dict, dict: initialized distance and path dictionaries
    """
    distance = {}
    path = {}
    for node in graph:
        distance[node] = float('Inf')
        path[node] = None
    distance[source] = 0
    return distance, path

def relax(u_key, v_key, graph, distance, path):
    """
    Checks if shorter path exists.

    Args:
        u (key): key of a node
        v (key): key of another node
        graph (dict): the graph object
        distance (dict): the distances dictionary
        path (dict): the paths dictionary
    """
    if distance[v_key] > distance[u_key] + graph[u_key][v_key]:
        distance[v_key] = distance[u_key] + graph[u_key][v_key]
        path[v_key] = u_key

def bellman_ford(graph, source):
    """
    Main entry point of the Bellman-Ford algorithm

    Args:
        graph (dict): a graph in dictionary representation
        source (key): the key of the source node
    """
    distance, path = initialize(graph, source)
    for _ in range(len(graph)-1):
        for u_key in graph:
            for v_key in graph[u_key]:
                relax(u_key, v_key, graph, distance, path)
    for u_key in graph:
        for v_key in graph[u_key]:
            assert distance[v_key] <= distance[u_key] + graph[u_key][v_key]
    return distance, path


class DBSMOTE(OverSampling):
    """
    References:
        * BibTex::

            @Article{dbsmote,
                        author="Bunkhumpornpat, Chumphol
                        and Sinapiromsaran, Krung
                        and Lursinsap, Chidchanok",
                        title="DBSMOTE: Density-Based Synthetic Minority
                                Over-sampling TEchnique",
                        journal="Applied Intelligence",
                        year="2012",
                        month="Apr",
                        day="01",
                        volume="36",
                        number="3",
                        pages="664--684",
                        issn="1573-7497",
                        doi="10.1007/s10489-011-0287-y",
                        url="https://doi.org/10.1007/s10489-011-0287-y"
                        }

    Notes:
        * Standardization is needed to use absolute eps values.
        * The clustering is likely to identify all instances as noise, fixed
            by recursive call with increaseing eps.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_uses_clustering,
                  OverSampling.cat_density_based,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 eps=0.8,
                 min_samples=3,
                 db_iter_limit=3,
                 nn_params={},
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            eps (float): eps parameter of DBSCAN
            min_samples (int): min_samples parameter of DBSCAN
            db_iter_limit (int): DBSCAN iterations limit
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater(eps, "eps", 0)
        self.check_greater_or_equal(min_samples, "min_samples", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.eps = eps
        self.min_samples = min_samples
        self.db_iter_limit = db_iter_limit
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'eps': [0.5, 0.8, 1.2],
                                  'min_samples': [1, 3, 5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def execute_dbscan(self, X_min):
        """
        Execute DBSCAN by releasing thresholds

        Args:
            X_min (np.array): the minority samples

        Returns:
            np.array: the labels
        """
        dbscan = DBSCAN(eps=self.eps,
                    min_samples=self.min_samples,
                    n_jobs=self.n_jobs).fit(X_min)
        labels = dbscan.labels_
        num_labels = np.max(labels) + 1

        step = 1

        while num_labels == 0 and step < self.db_iter_limit:
            # adjusting the parameters if no clusters were identified
            dbscan = DBSCAN(eps=self.eps*1.5**step,
                    min_samples=self.min_samples - 1*step,
                    n_jobs=self.n_jobs).fit(X_min)
            labels = dbscan.labels_
            num_labels = np.max(labels) + 1
            step = step + 1

        return labels

    def extract_shortest_paths(self,
                                X_min,
                                clusters,
                                nn_params):
        """
        Extract shortest paths.

        Args:
            X_min (np.array): minority samples
            clusters (np.array): clusters
            nn_params (dict): nearest neighbors params

        Returns:
            list: shortest paths
        """
        shortest_paths = []
        for _, clust in enumerate(clusters):
            # extracting the cluster elements
            cluster = X_min[clust]
            # initializing the graph object
            graph = {}
            for idx in range(len(cluster)):
                graph[idx] = {}

            # fitting nearest neighbors model to the cluster elements
            nnmt = NearestNeighborsWithMetricTensor(n_neighbors=len(cluster),
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
            nnmt.fit(cluster)
            dist, ind = nnmt.kneighbors(cluster)

            # extracting graph edges according directly to the density reachabality
            # definition
            for idx in range(len(cluster)):
                index_set = ind[idx][1:min([len(cluster),
                                            (self.min_samples + 1)])]
                for jdx in range(len(cluster)):
                    if jdx in index_set \
                        and dist[idx][ind[idx] == jdx][0] < self.eps:
                        graph[idx][jdx] = dist[idx][ind[idx] == jdx][0]

            # finding the index of the center like object
            centroid_ind = nnmt.kneighbors(np.mean(cluster, axis=0).reshape(1, -1),
                                           return_distance=False)[0][0]

            # extracting shortest paths from centroid object
            shortest_paths.append(bellman_ford(graph, centroid_ind))

        return shortest_paths

    def generate_samples(self,
                            n_to_sample,
                            clusters,
                            shortest_paths,
                            X_min):
        """
        Generate samples

        Args:
            n_to_sample (int): number of samples to be generates
            clusters (list): clusters
            shortest_paths (list): the shortest paths
            X_min (np.array): the minority samples

        Returns:
            np.array: the generated samples
        """
        cluster_dist = np.array([len(clusters[idx])
                                  for idx in range(len(clusters))])
        cluster_dist = cluster_dist/np.sum(cluster_dist)

        # generating samples
        samples = []
        while len(samples) < n_to_sample:
            cluster_idx = self.random_state.choice(
                np.arange(len(clusters)), p=cluster_dist)
            cluster = X_min[clusters[cluster_idx]]
            idx = self.random_state.choice(range(len(clusters[cluster_idx])))

            # executing shortest path algorithm
            _, parents = shortest_paths[cluster_idx]

            # extracting path
            path = [idx]
            while not parents[path[-1]] is None:
                path.append(parents[path[-1]])

            if len(path) == 1:
                # if the center like object is selected
                samples.append(cluster[path[0]])
            elif len(path) == 2:
                # if the path consists of 1 edge
                sample = \
                    self.sample_between_points_componentwise(cluster[path[0]],
                                                             cluster[path[1]])
                samples.append(sample)
            else:
                # if the path consists of at least two edges
                r_vertex = self.random_state.randint(len(path)-1)
                sample = \
                    self.sample_between_points_componentwise(cluster[path[r_vertex]],
                                                             cluster[path[r_vertex + 1]])
                samples.append(sample)
        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        scaler = StandardScaler().fit(X)
        X_ss = scaler.transform(X) # pylint: disable=invalid-name

        # doing the clustering using DBSCAN
        X_min = X_ss[y == self.min_label]
        labels = self.execute_dbscan(X_min)
        num_labels = np.max(labels) + 1

        if num_labels == 0:
            return self.return_copies(X, y, "no reasonable clusters found")

        # determining cluster size distribution
        clusters = [np.where(labels == i)[0] for i in range(num_labels)]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        shortest_paths = self.extract_shortest_paths(X_min, clusters, nn_params)

        samples = self.generate_samples(n_to_sample, clusters, shortest_paths,
                            X_min)

        return (np.vstack([X, scaler.inverse_transform(np.vstack(samples))]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'eps': self.eps,
                'min_samples': self.min_samples,
                'db_iter_limit': self.db_iter_limit,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing oversampling and noise removal with high dimensionality data.
"""

import numpy as np
import pytest

import smote_variants as sv

from smote_variants.datasets import load_high_dim
from .additional_objs import additional_objs

dataset = load_high_dim()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_high_dim(smote_class):
    """
    Testing an oversampler with high dimensionality data.

    Args:
        smote_class (class): an oversampler class
    """
    smote_obj = smote_class()
    X, y = smote_obj.sample(dataset['data'], dataset['target'])
    assert len(X) > 0
    assert X.shape[1] == smote_obj\
                                .preprocessing_transform(dataset['data']).shape[1]
    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", additional_objs)
def test_high_dim_plus(smote_obj):
    """
    Testing some specific parameterizations with high dimensionality data.

    Args:
        smote_obj (obj): an oversampling object
    """
    X, y = smote_obj.sample(dataset['data'], dataset['target'])
    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("nf_class", sv.get_all_noisefilters())
def test_normal(nf_class):
    """
    Testing a noise filter with high dimensionality data

    Args:
        nf_class (class): a noise filter class
    """
    X, y = nf_class().remove_noise(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
This module implements the SMOTE_RSB method.
"""

import numpy as np

from sklearn.metrics import pairwise_distances

from ..base import coalesce, coalesce_dict
from ..base import OverSampling
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['SMOTE_RSB']

class SMOTE_RSB(OverSampling):
    """
    References:
        * BibTex::

            @Article{smote_rsb,
                    author="Ramentol, Enislay
                    and Caballero, Yail{\'e}
                    and Bello, Rafael
                    and Herrera, Francisco",
                    title="SMOTE-RSB*: a hybrid preprocessing approach
                            based on oversampling and undersampling for
                            high imbalanced data-sets using SMOTE and
                            rough sets theory",
                    journal="Knowledge and Information Systems",
                    year="2012",
                    month="Nov",
                    day="01",
                    volume="33",
                    number="2",
                    pages="245--265",
                    issn="0219-3116",
                    doi="10.1007/s10115-011-0465-6",
                    url="https://doi.org/10.1007/s10115-011-0465-6"
                    }

    Notes:
        * I think the description of the algorithm in Fig 5 of the paper
            is not correct. The set "resultSet" is initialized with the
            original instances, and then the While loop in the Algorithm
            run until resultSet is empty, which never holds. Also, the
            resultSet is only extended in the loop. Our implementation
            is changed in the following way: we generate twice as many
            instances are required to balance the dataset, and repeat
            the loop until the number of new samples added to the training
            set is enough to balance the dataset.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_ordinary,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors in the SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def normalization_factor(self, X_samp):
        """
        Construct the normalization factors.

        Args:
            X_samp (np.array): the generated samples

        Returns:
            np.array: the normalization factors
        """
        # Step 3: first the data is normalized
        maximums = np.max(X_samp, axis=0)
        minimums = np.min(X_samp, axis=0)

        # normalize X_new and X_maj
        norm_factor = maximums - minimums
        null_mask = norm_factor == 0
        n_null = np.sum(null_mask)
        fixed = np.max(np.vstack([maximums[null_mask], np.repeat(1, n_null)]),
                       axis=0)

        norm_factor[null_mask] = fixed

        return norm_factor

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_maj = X[y == self.maj_label]
        X_min = X[y == self.min_label]

        # Step 1: do the sampling
        smote = SMOTE(proportion=self.proportion,
                      n_neighbors=self.n_neighbors,
                      nn_params=self.nn_params,
                      ss_params=self.ss_params,
                      n_jobs=self.n_jobs,
                      random_state=self._random_state_init)

        X_samp, _ = smote.sample(X, y)
        X_samp = X_samp[len(X):]

        if len(X_samp) == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # Step 2: (original will be added later)
        result_set = np.zeros(shape=(0, X.shape[1]))

        norm_factor = self.normalization_factor(X_samp)

        # compute similarity matrix
        similarity_matrix = 1.0 - pairwise_distances(X_samp/norm_factor,
                                                     X_maj/norm_factor,
                                                     metric='minkowski',
                                                     p=1) / X.shape[1]

        # Step 4: counting the similar examples
        similarity_value = 0.4

        already_added = np.repeat(False, len(X_samp))
        while (result_set.shape[0] < X_maj.shape[0] - X_min.shape[0] \
                                            and similarity_value <= 0.9):
            conts = np.sum(similarity_matrix > similarity_value, axis=1)
            mask = (conts == 0) & (~ already_added)
            result_set = np.vstack([result_set, X_samp[mask]])
            already_added[mask] = True

            similarity_value = similarity_value + 0.05

        result_set = result_set[:np.min([result_set.shape[0],
                                        X_maj.shape[0] - X_min.shape[0]])]

        # Step 5: returning the results depending the number of instances
        # added to the result set
        return (np.vstack([X, result_set]),
                np.hstack([y, np.repeat(self.min_label,
                                        len(result_set))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the IPADE-ID.
"""
import numpy as np

import pytest

from smote_variants import IPADE_ID

def test_specific():
    """
    Oversampler specific testing
    """
    obj = IPADE_ID()

    with pytest.raises(Exception):
        obj.configuration_checks(y=np.array([0, 0]),
                                    for_validation=np.array([0, 1]),
                                    GS_y=None)
    with pytest.raises(Exception):
        obj.configuration_checks(y=np.array([0, 1]),
                                    for_validation=np.array([0, 1]),
                                    GS_y=np.array([0, 0]))

    X = np.array([[1, 1], [2, 1], [3, 1],
                    [4, 1], [5, 1], [6, 1],
                    [7, 1], [8, 1], [9, 1],
                    [10, 1], [11, 1]])
    y = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0])

    assert IPADE_ID().sample(X, y)[0].shape[0] == X.shape[0]



"""
This module implements the polynom_fit_SMOTE method.
"""

import numpy as np

from ..base import coalesce_dict
from ..base import OverSampling, OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['polynom_fit_SMOTE_star',
            'polynom_fit_SMOTE_bus',
            'polynom_fit_SMOTE_poly',
            'polynom_fit_SMOTE_mesh']

class polynom_fit_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{polynomial_fit_smote,
                            author={Gazzah, S. and Amara, N. E. B.},
                            booktitle={2008 The Eighth IAPR International
                                        Workshop on Document Analysis Systems},
                            title={New Oversampling Approaches Based on
                                    Polynomial Fitting for Imbalanced Data
                                    Sets},
                            year={2008},
                            volume={},
                            number={},
                            pages={677-684},
                            doi={10.1109/DAS.2008.74},
                            ISSN={},
                            month={Sept},}
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 *,
                 topology='star',
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            topoplogy (str): 'star'/'bus'/'poly_N'
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0.0)
        if topology.startswith('poly'):
            self.check_greater_or_equal(
                int(topology.split('_')[-1]), 'topology', 1)
        else:
            self.check_isin(topology, "topology", ['star', 'bus'])

        self.proportion = proportion
        self.topology = topology

    #@ classmethod
    #def parameter_combinations(cls, raw=False):
    #    """
    #    Generates reasonable parameter combinations.
    #
    #    Returns:
    #        list(dict): a list of meaningful parameter combinations
    #    """
    #    parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
    #                                             1.0, 1.5, 2.0],
    #                              'topology': ['star', 'bus',
    #                                           'poly_1', 'poly_2', 'poly_3']}
    #    return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # extracting minority samples
        X_min = X[y == self.min_label]

        samples = []
        if self.topology == 'star':
            # Implementation of the star topology
            kdx = np.max([1, int(np.rint(n_to_sample / X_min.shape[0]))])
            splits = np.arange(1, kdx + 1).astype(float)/(kdx + 1)

            X_mean = np.mean(X_min, axis=0) # pylint: disable=invalid-name
            diffs = X_mean - X_min

            samples = np.vstack(diffs[:, None] * splits[:, None])
            samples = samples + np.repeat(X_min,
                                            np.repeat(kdx, X_min.shape[0]),
                                            axis=0)
        elif self.topology == 'bus':
            # Implementation of the bus topology
            kdx = np.max([1, int(np.rint(n_to_sample / X_min.shape[0]))])
            splits = np.arange(1, kdx + 1).astype(float)/(kdx + 1)

            diffs = np.diff(X_min, axis=0)

            samples = np.vstack(diffs[:, None] * splits[:, None])

            samples = samples + np.repeat(X_min[:-1],
                                            np.repeat(kdx, X_min.shape[0] - 1),
                                            axis=0)
        elif self.topology.startswith('poly'):
            # Implementation of the polynomial topology
            degree = int(self.topology.split('_')[1])

            # this hack is added to make the fitted polynoms independent
            # from the ordering of the minority samples
            X_min = X_min[np.mean(X_min, axis=1).argsort()]

            def fit_poly(dim):
                return np.poly1d(np.polyfit(np.arange(len(X_min)),
                                            X_min[:, dim],
                                            degree))

            polys = [fit_poly(dim) for dim in range(X_min.shape[1])]

            for dim in range(X_min.shape[1]):
                rand = self.random_state.random_sample(size=n_to_sample)\
                                                                * X_min.shape[0]
                samples.append(np.array(polys[dim](rand)))

            samples = np.vstack(samples).T

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'topology': self.topology,
                **OverSampling.get_params(self)}

class polynom_fit_SMOTE_star(OverSampling):# pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{polynomial_fit_smote,
                            author={Gazzah, S. and Amara, N. E. B.},
                            booktitle={2008 The Eighth IAPR International
                                        Workshop on Document Analysis Systems},
                            title={New Oversampling Approaches Based on
                                    Polynomial Fitting for Imbalanced Data
                                    Sets},
                            year={2008},
                            volume={},
                            number={},
                            pages={677-684},
                            doi={10.1109/DAS.2008.74},
                            ISSN={},
                            month={Sept},}
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 *,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.polynom_fit_smote = polynom_fit_SMOTE(proportion=proportion,
                                                    topology='star',
                                                    random_state=random_state)

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        params = self.polynom_fit_smote.get_params()
        params ['class_name'] = self.__class__.__name__
        return params

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        return self.polynom_fit_smote.sample(X, y)

class polynom_fit_SMOTE_bus(OverSampling):# pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{polynomial_fit_smote,
                            author={Gazzah, S. and Amara, N. E. B.},
                            booktitle={2008 The Eighth IAPR International
                                        Workshop on Document Analysis Systems},
                            title={New Oversampling Approaches Based on
                                    Polynomial Fitting for Imbalanced Data
                                    Sets},
                            year={2008},
                            volume={},
                            number={},
                            pages={677-684},
                            doi={10.1109/DAS.2008.74},
                            ISSN={},
                            month={Sept},}
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 *,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.polynom_fit_smote = polynom_fit_SMOTE(proportion=proportion,
                                                    topology='bus',
                                                    random_state=random_state)

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        params = self.polynom_fit_smote.get_params()
        params ['class_name'] = self.__class__.__name__
        return params

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        return self.polynom_fit_smote.sample(X, y)

class polynom_fit_SMOTE_poly(OverSampling):# pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{polynomial_fit_smote,
                            author={Gazzah, S. and Amara, N. E. B.},
                            booktitle={2008 The Eighth IAPR International
                                        Workshop on Document Analysis Systems},
                            title={New Oversampling Approaches Based on
                                    Polynomial Fitting for Imbalanced Data
                                    Sets},
                            year={2008},
                            volume={},
                            number={},
                            pages={677-684},
                            doi={10.1109/DAS.2008.74},
                            ISSN={},
                            month={Sept},}
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 order=2,
                 *,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            order (int): the order of the fitted polynoms
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.polynom_fit_smote = polynom_fit_SMOTE(proportion=proportion,
                                                    topology='poly_' + str(order),
                                                    random_state=random_state)
        self.order = order

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'order': [1, 2, 3]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        params = {'order': self.order,
                **self.polynom_fit_smote.get_params()}
        params['class_name'] = self.__class__.__name__
        return params

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        return self.polynom_fit_smote.sample(X, y)

class polynom_fit_SMOTE_mesh(OverSamplingSimplex): # pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{polynomial_fit_smote,
                            author={Gazzah, S. and Amara, N. E. B.},
                            booktitle={2008 The Eighth IAPR International
                                        Workshop on Document Analysis Systems},
                            title={New Oversampling Approaches Based on
                                    Polynomial Fitting for Imbalanced Data
                                    Sets},
                            year={2008},
                            volume={},
                            number={},
                            pages={677-684},
                            doi={10.1109/DAS.2008.74},
                            ISSN={},
                            month={Sept},}
    """

    categories = [OverSamplingSimplex.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 ss_params=None,
                 *,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            ss_params (dict): simplex sampling parameters
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'deterministic',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0.0)

        self.proportion = proportion

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # extracting minority samples
        X_min = X[y == self.min_label]

        kdx = np.max([1, int(np.rint(n_to_sample / X_min.shape[0]))])

        samples = []

        base = np.arange(X_min.shape[0])
        neighbors = np.vstack([self.random_state.choice(np.arange(X_min.shape[0]),
                                        X_min.shape[0],
                                        replace=True) for _ in range(kdx)])

        indices = np.vstack([base, neighbors]).T

        samples = self.sample_simplex(X=X_min,
                                        indices=indices,
                                        n_to_sample=n_to_sample)


        ## Implementation of the mesh topology
        #if len(X_min)**2 > n_to_sample:
        #    while len(samples) < n_to_sample:
        #        random_i = self.random_state.randint(len(X_min))
        #        random_j = self.random_state.randint(len(X_min))
        #        diff = X_min[random_i] - X_min[random_j]
        #        samples.append(X_min[random_j] + 0.5*diff)
        #else:
        #    n_combs = (len(X_min)*(len(X_min)-1)/2)
        #    k = max([1, int(np.rint(n_to_sample/n_combs))])
        #    for i in range(len(X_min)):
        #        for j in range(len(X_min)):
        #            diff = X_min[i] - X_min[j]
        #            for li in range(1, k+1):
        #                samples.append(X_min[j] + float(li)/(k+1)*diff)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Edge_Det_SMOTE method.
"""

import numpy as np

from ..base import fix_density, coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['Edge_Det_SMOTE']

class Edge_Det_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{Edge_Det_SMOTE,
                            author={Kang, Y. and Won, S.},
                            booktitle={ICCAS 2010},
                            title={Weight decision algorithm for oversampling
                                    technique on class-imbalanced learning},
                            year={2010},
                            volume={},
                            number={},
                            pages={182-186},
                            keywords={edge detection;learning (artificial
                                        intelligence);weight decision
                                        algorithm;oversampling technique;
                                        class-imbalanced learning;class
                                        imbalanced data problem;edge
                                        detection algorithm;spatial space
                                        representation;Classification
                                        algorithms;Image edge detection;
                                        Training;Noise measurement;Glass;
                                        Training data;Machine learning;
                                        Imbalanced learning;Classification;
                                        Weight decision;Oversampling;
                                        Edge detection},
                            doi={10.1109/ICCAS.2010.5669889},
                            ISSN={},
                            month={Oct}}

    Notes:
        * This technique is very loosely specified.
    """

    categories = [OverSamplingSimplex.cat_density_based,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 k=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            k (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(k, "k", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.k = k # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'k': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_magnitudes(self, X, y):
        """
        Determine the magnitudes.

        Args:
            X (np.array): the feature vectors
            y (np.array): the targets

        Returns:
            np.array: the magnitudes of minority points
        """
        # organizing class labels according to feature ranking
        magnitudes = np.zeros(X.shape[0])
        for idx in range(X.shape[1]):
            to_sort = zip(X[:, idx], np.arange(len(X)), y)
            _, ind, label = zip(*sorted(to_sort, key=lambda x: x[0]))
            # extracting edge magnitudes in this dimension
            for jdx in range(1, len(ind)-1):
                magnitudes[ind[jdx]] = magnitudes[ind[jdx]] + \
                    (label[jdx-1] - label[jdx+1])**2

        # density estimation
        magnitudes = magnitudes[y == self.min_label]
        magnitudes[magnitudes < 0]= 0
        magnitudes = fix_density(magnitudes)
        magnitudes = np.sqrt(magnitudes)
        magnitudes = magnitudes/np.sum(magnitudes)

        return magnitudes

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        magnitudes = self.determine_magnitudes(X, y)

        # fitting nearest neighbors models to minority samples
        n_neighbors = min([len(X_min), self.k+1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        # do the sampling

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        base_weights=magnitudes)

        #samples = []
        #for _ in range(n_to_sample):
        #    idx = self.random_state.choice(np.arange(len(X_min)), p=magnitudes)
        #    X_a = X_min[idx]
        #    X_b = X_min[self.random_state.choice(ind[idx][1:])]
        #    samples.append(self.sample_between_points(X_a, X_b))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'k': self.k,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the Selected_SMOTE method.
"""

import numpy as np

from ..base import coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['Selected_SMOTE']

class Selected_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_out_smote_cosine_selected_smote,
                    title={SMOTE-Out, SMOTE-Cosine, and Selected-SMOTE: An
                                enhancement strategy to handle imbalance in
                                data level},
                    author={Fajri Koto},
                    journal={2014 International Conference on Advanced
                                Computer Science and Information System},
                    year={2014},
                    pages={280-284}
                    }

    Notes:
        * Significant attribute selection was not described in the paper,
            therefore we have implemented something meaningful.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_componentwise,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 perc_sign_attr=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            strategy (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): parameter of the NearestNeighbors component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            perc_sign_attr (float): [0,1] - percentage of significant
                                            attributes
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_in_range(perc_sign_attr, 'perc_sign_attr', [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.perc_sign_attr = perc_sign_attr
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'perc_sign_attr': [0.3, 0.5, 0.8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def significant_attributes(self, X_min, X_maj):
        """
        Significant attribute selection.

        Args:
            X_min (np.array): the minority samples
            X_maj (np.array): the majority samples

        Returns:
            np.array: the significant attribute mask
        """
        # significant attribute selection was not described in the paper
        # I have implemented significant attribute selection by checking
        # the overlap between ranges of minority and majority class attributes
        # the attributes with bigger overlap respecting their ranges
        # are considered more significant
        min_ranges_a = np.min(X_min, axis=0)
        min_ranges_b = np.max(X_min, axis=0)
        maj_ranges_a = np.min(X_maj, axis=0)
        maj_ranges_b = np.max(X_maj, axis=0)

        # end points of overlaps
        max_a = np.max(np.vstack([min_ranges_a, maj_ranges_a]), axis=0)
        min_b = np.min(np.vstack([min_ranges_b, maj_ranges_b]), axis=0)

        # size of overlap
        overlap = min_b - max_a

        # replacing negative values (no overlap) by zero
        overlap = np.where(overlap < 0, 0, overlap)
        # percentage of overlap compared to the ranges of attributes in the
        # minority set
        percentages = min_ranges_b - min_ranges_a
        percentages[percentages != 0] = overlap[percentages != 0] \
                                            / percentages[percentages != 0]

        # number of significant attributes to determine
        tmp = int(np.rint(self.perc_sign_attr*len(percentages)))
        num_sign_attr = np.min([1, tmp])

        sign_attr = (percentages >= sorted(percentages)[-num_sign_attr])
        sign_attr = sign_attr.astype(int)

        return sign_attr

    def generate_samples(self,
                            X_min,
                            base_indices,
                            neigh_indices,
                            significant_attr):
        """
        Generate the samples.

        Args:
            X_min (np.array): minority samples
            base_indices (np.array): indices of base vectors
            neigh_indices (np.array): indices of neighbor vectors
            significant_attr (np.array): significant attribute mask

        Returns:
            np.array: the generated samples
        """
        base_vectors = X_min[base_indices]
        neigh_vectors = X_min[neigh_indices]
        random = self.random_state.random_sample(size=base_vectors.shape)
        samples = base_vectors + \
                    (neigh_vectors - base_vectors) * random * significant_attr
        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)
        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        n_neighbors = np.min([len(X_min), self.n_neighbors + 1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        **(nn_params))
        nnmt.fit(X_min)

        indices = nnmt.kneighbors(X_min, return_distance=False)

        significant_attr = self.significant_attributes(X_min,
                                                        X[y == self.maj_label])

        base_indices = self.random_state.choice(np.arange(X_min.shape[0]),
                                                n_to_sample)
        neigh_indices = self.random_state.choice(np.arange(1, n_neighbors),
                                                    n_to_sample)

        samples = self.generate_samples(X_min=X_min,
                                        base_indices=base_indices,
                                        neigh_indices=indices[base_indices,
                                                            neigh_indices],
                                        significant_attr=significant_attr)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'perc_sign_attr': self.perc_sign_attr,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the queries.
"""

import smote_variants as sv

def test_queries():
    """
    Testing the queries.
    """
    assert len(sv.get_all_oversamplers()) > 0
    assert len(sv.get_all_oversamplers(n_quickest=5)) == 5
    assert len(sv.get_all_noisefilters()) > 0
    assert len(sv.get_multiclass_oversamplers()) > 0
    assert len(sv.get_multiclass_oversamplers(n_quickest=5)) == 5
    assert len(sv.get_metric_learning_oversamplers()) > 0
    assert len(sv.get_metric_learning_oversamplers(n_quickest=5)) == 5
    assert len(sv.get_simplex_sampling_oversamplers()) > 0
    assert len(sv.get_simplex_sampling_oversamplers(n_quickest=5)) == 5
    assert len(sv.get_simplex_sampling_oversamplers(
                            within_simplex_sampling='random')) > 0
    assert len(sv.get_simplex_sampling_oversamplers(
                            within_simplex_sampling=['random'])) > 0
    assert len(sv.get_simplex_sampling_oversamplers(
                        exclude_within_simplex_sampling='deterministic')) > 0
    assert len(sv.get_simplex_sampling_oversamplers(
                        exclude_within_simplex_sampling=['deterministic'])) > 0
    assert len(sv.get_simplex_sampling_oversamplers(
                                            with_gaussian_component=True)) > 0
    assert len(sv.get_simplex_sampling_oversamplers(
                                            with_gaussian_component=False)) > 0
    assert len(sv.get_simplex_sampling_oversamplers(n_dim_range=2)) > 0
    assert len(sv.get_simplex_sampling_oversamplers(n_dim_range=[1, 4])) > 0

    oversamplers = sv.get_all_oversamplers()[:5]
    assert len(sv.generate_parameter_combinations(oversamplers,
                                                result_format='dict')) > 0

    assert len(sv.generate_parameter_combinations(oversamplers,
                                        result_format='smote_variants')) > 0



"""
Testing the VIS_RST.
"""
import numpy as np

import pytest

from smote_variants import VIS_RST

def test_specific():
    """
    Oversampler specific testing
    """
    obj = VIS_RST()

    assert len(obj.parameter_combinations(raw=True)) > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9], [9, 10], [10, 11],
                    [11, 12], [12, 13], [13, 14], [14, 15],
                    [15, 16], [16, 17]])
    y = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    X = np.array([[1, 2], [1, 2], [1, 2], [1, 2], [1, 2], [1, 2],
                    [3, 4], [3, 4], [3, 4], [3, 4], [3, 4]])
    y = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

    assert obj.sample(X, y)[0].shape[0] > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9], [9, 10], [10, 11],
                    [11, 12], [12, 13], [13, 14], [14, 15],
                    [15, 16], [16, 17]])
    y = np.array([0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    with pytest.raises(ValueError):
        obj.check_all_boundary(np.array([1]))

    with pytest.raises(ValueError):
        obj.check_labels(np.array(['NOI']))

    with pytest.raises(ValueError):
        obj.check_all_noise(np.array([1]))

    assert obj.set_mode(None, np.array(['DAN'])) == 'no_safe'

    assert obj.set_mode(np.zeros((0, 2)), np.array(['DAN', 'SAF'])) == 'high_complexity'

    ind = [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]
    labels = np.array(['SAF', 'DAN', 'NOI'])[ind]

    assert len(obj.generate_samples(X_min=X,
                                    nn_params={},
                                    n_to_sample=10,
                                    labels=labels,
                                    mode='high_complexity')) > 0

    assert len(obj.generate_samples(X_min=X,
                                    nn_params={},
                                    n_to_sample=10,
                                    labels=labels,
                                    mode='low_complexity')) > 0

    assert len(obj.generate_samples(X_min=X,
                                    nn_params={},
                                    n_to_sample=10,
                                    labels=labels,
                                    mode='other')) > 0



"""
Testing the dataset folding.
"""

import os
import shutil

from sklearn import datasets
import pytest

from smote_variants.evaluation import Folding

dataset = datasets.load_breast_cancer()
dataset['name'] = 'breast_cancer'

cache_path = os.path.join('.', 'test_path')

shutil.rmtree(cache_path, ignore_errors=True)

def test_basic():
    """
    Testing the basic functionalities.
    """
    folding = Folding(dataset,
                        cache_path=cache_path,
                        serialization='json')
    lens = [len(fold['X_test']) for fold in folding.fold()]

    folding.cache_foldings()

    assert len(lens) == 10
    assert len(folding.folding_files()) == 10

def test_json_specific_validator():
    """
    Testing the json serialization.
    """
    folding = Folding(dataset,
                        validator_params={'n_repeats': 1,
                                            'n_splits': 5,
                                            'random_state': 5},
                        cache_path=cache_path,
                        serialization='json',
                        reset=True)
    lens = [len(fold['X_test']) for fold in folding.fold()]
    assert len(lens) == 5

    folding = Folding(dataset,
                        cache_path=cache_path,
                        serialization='json',
                        reset=True)
    lens = [len(fold['X_test']) for fold in folding.fold()]
    assert len(lens) == 10

def test_pickle():
    """
    Testing the pickle serialization.
    """
    folding = Folding(dataset,
                        cache_path=cache_path,
                        serialization='pickle')
    lens = [len(fold['X_test']) for fold in folding.fold()]
    assert len(lens) == 10
    assert len(folding.folding_files()) == 10

def test_nonsense():
    """
    Testing with a nonsense parameter.
    """
    with pytest.raises(ValueError):
        Folding(dataset,
                cache_path=cache_path,
                serialization='nonsense').cache_foldings()

def test_get_params():
    """
    Testing the get_params function.
    """
    folding = Folding(dataset,
                        cache_path=cache_path,
                        serialization='json')
    assert str(folding.get_params()) == folding.descriptor()



"""
This module implements the edited nearest neighbors technique.
"""
import numpy as np

from ..base import mode, coalesce
from ._noisefilter import NoiseFilter
from ..base import (NearestNeighborsWithMetricTensor)

from .._logger import logger
_logger= logger

__all__= ['EditedNearestNeighbors']

class EditedNearestNeighbors(NoiseFilter):
    """
    References:
        * BibTex::

            @article{smoteNoise0,
                     author = {Batista, Gustavo E. A. P. A. and Prati,
                                Ronaldo C. and Monard, Maria Carolina},
                     title = {A Study of the Behavior of Several Methods for
                                Balancing Machine Learning Training Data},
                     journal = {SIGKDD Explor. Newsl.},
                     issue_date = {June 2004},
                     volume = {6},
                     number = {1},
                     month = jun,
                     year = {2004},
                     issn = {1931-0145},
                     pages = {20--29},
                     numpages = {10},
                     url = {http://doi.acm.org/10.1145/1007730.1007735},
                     doi = {10.1145/1007730.1007735},
                     acmid = {1007735},
                     publisher = {ACM},
                     address = {New York, NY, USA}
                    }
    """

    def __init__(self,
                 remove='both',
                 nn_params=None,
                 n_jobs=1,
                 **_kwargs):
        """
        Constructor of the noise removal object

        Args:
            remove (str): class to remove from 'both'/'min'/'maj'
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
        """
        super().__init__()

        self.check_isin(remove, 'remove', ['both', 'min', 'maj'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.remove = remove
        self.nn_params= coalesce(nn_params, {})
        self.n_jobs = n_jobs

    def get_params(self, deep=False):
        return {'remove': self.remove,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **NoiseFilter.get_params(self, deep)}

    def remove_noise(self, X, y):
        """
        Removes noise

        Args:
            X (np.array): features
            y (np.array): target labels

        Returns:
            np.array, np.array: cleaned features and target labels
        """
        _logger.info("%s: Running noise removal", self.__class__.__name__)
        self.class_label_statistics(y)

        if len(X) < 4:
            _logger.info("%s: Not enough samples for noise removal",
                                                self.__class__.__name__)
            return X.copy(), y.copy()

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=4,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        indices= nnmt.fit(X).kneighbors(X, return_distance=False)

        to_remove = []
        for idx in range(len(X)):
            if not y[idx] == mode(y[indices[idx][1:]]):
                if (self.remove == 'both' or
                    (self.remove == 'min' and y[idx] == self.min_label) or
                        (self.remove == 'maj' and y[idx] == self.maj_label)):
                    to_remove.append(idx)

        return np.delete(X, to_remove, axis=0), np.delete(y, to_remove)



"""
This module implements the cluster_SMOTE method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans

from ..config import suppress_external_warnings
from ..base import coalesce_dict, coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['cluster_SMOTE']

class cluster_SMOTE(OverSamplingSimplex): # pylint: disable=invalid-name
    """
    References:
        * BibTex::

            @INPROCEEDINGS{cluster_SMOTE,
                            author={Cieslak, D. A. and Chawla, N. V. and
                                        Striegel, A.},
                            booktitle={2006 IEEE International Conference
                                        on Granular Computing},
                            title={Combating imbalance in network
                                        intrusion datasets},
                            year={2006},
                            volume={},
                            number={},
                            pages={732-737},
                            keywords={Intelligent networks;Intrusion detection;
                                        Telecommunication traffic;Data mining;
                                        Computer networks;Data security;
                                        Machine learning;Counting circuits;
                                        Computer security;Humans},
                            doi={10.1109/GRC.2006.1635905},
                            ISSN={},
                            month={May}}
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=3,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_clusters=3,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in SMOTE
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): the simplex sampling parameters
            n_clusters (int): number of clusters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_clusters, "n_components", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_clusters = n_clusters
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_clusters': [3, 5, 7, 9]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_clusters_and_neighbood_graph(self,
                                                cluster_indices,
                                                X_min,
                                                nn_params):
        """
        Determine clusters and neighborhood graphs.

        Args:
            cluster_indices (np.array): the cluster labeling
            X_min (np.array): the vectors
            nn_params (dict): nearest neighbors parameters

        Returns:
            np.array, np.array: the clusters and the neighborhood graphs
        """
        cluster_vectors = [X_min[cluster_indices[idx]] \
                                for idx in range(len(cluster_indices))]
        cluster_nn_indices = []

        for idx, cluster in enumerate(cluster_vectors):
            n_neighbors = min([self.n_neighbors, len(cluster_indices[idx])])
            nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
            nnmt.fit(cluster)
            cluster_nn_indices.append(nnmt.kneighbors(cluster,
                                                        return_distance=False))

        return cluster_vectors, cluster_nn_indices

    def generate_samples_in_clusters(self,
                                    cluster_indices,
                                    X_min,
                                    nn_params,
                                    n_to_sample):
        """
        Generate samples within the clusters.

        Args:
            cluster_indices (np.array): the cluster labeling
            X_min (np.array): the vectors
            nn_params (dict): the nearest neighbors parameters
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        clusters, indices = \
            self.determine_clusters_and_neighbood_graph(cluster_indices,
                                                        X_min,
                                                        nn_params)

        clusters_selected = self.random_state.choice(len(clusters),
                                                        n_to_sample)
        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)

        samples = []
        for idx, cluster in enumerate(cluster_unique):
            #if len(clusters[cluster]) >= self.n_dim:
            samples.append(self.sample_simplex(X=clusters[cluster],
                                        indices=indices[cluster],
                                        n_to_sample=cluster_count[idx]))
            #else:
            #    sample_indices = self.random_state.choice(len(clusters[cluster]),
            #                                                cluster_count[idx])
            #    samples.append(clusters[cluster][sample_indices])

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determining the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        kmeans = KMeans(n_clusters=min([len(X_min), self.n_clusters]),
                        random_state=self._random_state_init)
        with warnings.catch_warnings():
            if suppress_external_warnings():
                warnings.simplefilter("ignore")
            kmeans.fit(X_min)

        unique_labels = np.unique(kmeans.labels_)

        # creating nearest neighbors objects for each cluster
        cluster_indices = [np.where(kmeans.labels_ == c)[0]
                           for c in unique_labels]

        if np.max([len(cluster) for cluster in cluster_indices]) <= 1:
            return self.return_copies(X, y, "All clusters contain 1 element")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        samples = self.generate_samples_in_clusters(cluster_indices,
                                                    X_min,
                                                    nn_params,
                                                    n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_clusters': self.n_clusters,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}
